/**
 * bipartiteneighbourhood.c
 *
 * A program which tests if Assumption 1 of Theorem 4.6 of [1] holds for the
 * inputgraphs which are read from stdin.
 * 
 * Note: the inputgraphs must be provided in multicode.
 * 
 * [1] M. Chudvnosky, J. Goedgebeur, O. Schaudt and M. Zhong, Obstructions for 
 * three-coloring graphs without induced paths on six vertices.
 * 
 * The latest version of BipartiteNeighbourhood can be found here:
 * http://caagt.ugent.be/tripods/
 * 
 * Author: Jan Goedgebeur (jan.goedgebeur@ugent.be)
 * In collaboration with: Maria Chudnovsky, Oliver Schaudt and Mingxian Zhong
 *
 * -----------------------------------
 * 
 * Copyright (c) 2015 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Compile with:
 *
 * cc -O4 bipartiteneighbourhood.c -o bipartiteneighbourhood
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include "util.h"

/*************************Defines and global variables*************************/

//The neighbourhood of each vertex (the vertex itself is not included in the neighbourhood)
static unsigned long long int neighbourhood_bitvector[MAXN];

static int vertex_found_where_property_doesnt_hold = 0;

static int number_of_bipartitions = 0;
static int number_of_reachable_bipartitions = 0;

unsigned char considered_vertex = 0;

typedef unsigned char EDGE[2];


/*********************************Methods**************************************/

/**
 * Returns 1 if all vertices in the set are nonadjacent, else returns 0.
 */
static int
vertices_in_set_are_nonadjacent(unsigned char set[], int set_size) {
    int i;
    unsigned long long int neighbours_of_set_vertices = 0;
    for(i = 0; i < set_size; i++)
        neighbours_of_set_vertices |= neighbourhood_bitvector[set[i]];
    
    for(i = 0; i < set_size; i++)
        if((BIT(set[i]) & neighbours_of_set_vertices) != 0)
            return 0; //set[i] is a neighbour of a vertex in the set!
    
    return 1;
}

/******************************************************************************/

/**
 * Returns 1 Assumption 1.(i) of Theorem 4.6 of [1] holds for the bipartition,
 * else returns 0.
 */
static int
neighbourhood_condition_holds(unsigned char partition_A[], int partition_A_size,
        unsigned char partition_B[], int partition_B_size,
        unsigned long long int forbidden_neighbourhood_vertices) {
    int i, j;
    for(i = 0; i < partition_A_size; i++)
        for(j = 0; j < partition_B_size; j++) {
            //Condition holds for pair (x,y) if:
            //(a) x and y are neighbours, or
            //(b) x has a neighbour in V(H) \ (N(v) + {v}), or
            //(c) y has a neighbour in V(H) \ (N(v) + {v})
            if((neighbourhood_bitvector[partition_A[i]] & BIT(partition_B[j])) == 0
                    && (neighbourhood_bitvector[partition_A[i]] & ~forbidden_neighbourhood_vertices) == 0
                    && (neighbourhood_bitvector[partition_B[j]] & ~forbidden_neighbourhood_vertices) == 0) {
                fprintf(stderr, "Info: Neighbourhood condition does not hold for x: %d and y: %d\n",
                        partition_A[i], partition_B[j]);
                return 0;
            }
        }
    
    return 1;
}

/******************************************************************************/

/**
 * Returns 1 if Assumption 1.(ii) of Theorem 4.6 of [1] holds, else returns 0.
 */
static int
path_condition_holds(unsigned char partition[], unsigned char partition_size,
        unsigned long long int forbidden_neighbourhood_vertices) {
    
    if(partition_size == 1)
        return 1;
    
    /*
     (A2_1)  for every x,x' in X there exists a  p  in V(H)  \   {v}  such that
         (4a) p is non-adjacent to v,
     * AND
         (4b) p is adjacent to exactly one of x,x'.
     */
    int i, j;
    //Number of combinations: (n choose 2)
    for(i = 0; i < partition_size; i++)
        for(j = i + 1; j < partition_size; j++) {
            int x = partition[i];
            int x_prime = partition[j];
            
            int correct_p_found = 0;
            int p;
            for(p = 0; p < number_of_vertices && !correct_p_found; p++)
                if(p != considered_vertex 
                        && (neighbourhood_bitvector[considered_vertex] & BIT(p)) == 0) {
                    int p_is_neighbour_of_x = (neighbourhood_bitvector[x] & BIT(p)) != 0;
                    int p_is_neighbour_of_x_prime = (neighbourhood_bitvector[x_prime] & BIT(p)) != 0;
                    if(p_is_neighbour_of_x_prime + p_is_neighbour_of_x == 1)
                        correct_p_found = 1;
                }
            
            if(!correct_p_found) {
                fprintf(stderr, "Info: condition A2_1 does not hold: No valid p found between %d and %d\n", x, x_prime);
                return 0;
            }
        }
    
    return 1;
}

/******************************************************************************/

/**
 * Returns 1 if Assumption 1 of Theorem 4.6 of [1] holds for this bipartition, 
 * else returns 0.
 */
static int
conditions_hold_for_bipartition(unsigned char partition_A[], int partition_A_size,
        unsigned char partition_B[], int partition_B_size,
        unsigned long long int forbidden_neighbourhood_vertices) {

    if(!neighbourhood_condition_holds(partition_A, partition_A_size,
        partition_B, partition_B_size, forbidden_neighbourhood_vertices))
        return 0;
    
    
    return path_condition_holds(partition_A, partition_A_size, forbidden_neighbourhood_vertices)
            && path_condition_holds(partition_B, partition_B_size, forbidden_neighbourhood_vertices);
    
}

/******************************************************************************/

/**
 * Makes all possible partitions of the vertices in neighbourhood and test if 
 * they are a valid bipartition (for the graph induced by them).
 */
//Max number of partitions is 2^(deg) - 1 since it is assumed that first element is fixed to avoid duplicates
static void
make_all_possible_bipartitions(unsigned char partition_A[], int partition_A_size,
        unsigned long long int partition_A_bitvector, 
        unsigned long long int forbidden_vertices_partition_A,
        unsigned char neighbourhood[], int neighbourhood_size, int current_index,
        unsigned long long int forbidden_neighbourhood_vertices) {
    
    if(partition_A_size == 0) {
        fprintf(stderr, "Error: partition cannot be empty!\n");
        exit(1);
    }

    if(partition_A_size < neighbourhood_size) {
        //Now test if B is a valid bipartition
        unsigned char partition_B[neighbourhood_size];
        int partition_B_size = 0;
        int i;
        for(i = 0; i < neighbourhood_size; i++)
            if((BIT(neighbourhood[i]) & partition_A_bitvector) == 0)
                partition_B[partition_B_size++] = neighbourhood[i];
        
        if(partition_A_size + partition_B_size != neighbourhood_size) {
            fprintf(stderr, "Error: invalid partition sizes!\n");
            exit(1);
        }
        
        //Test if B is a valid partition
        //If B is not valid, it could still become valid later since B becomes smaller
        if(vertices_in_set_are_nonadjacent(partition_B, partition_B_size)) {
            //Bipartition found...
            number_of_bipartitions++;
/*
            fprintf(stderr, "bipartition A: ");
            for(i = 0; i < partition_A_size; i++)
                fprintf(stderr, "%d ", partition_A[i]);
            fprintf(stderr, "\n");

            fprintf(stderr, "bipartition B: ");
            for(i = 0; i < partition_B_size; i++)
                fprintf(stderr, "%d ", partition_B[i]);
            fprintf(stderr, "\n");
            fprintf(stderr, "\n");
*/
            
            //Test if all pairs of vertices from partitions are reachable
            if(conditions_hold_for_bipartition(partition_A, partition_A_size,
                    partition_B, partition_B_size, forbidden_neighbourhood_vertices)) {
                //Could in principle already stop as soon as 1 valid one has been found
                number_of_reachable_bipartitions++;
            } else {
                fprintf(stderr, "The following bipartition is not reachable (with v=%d):\n", considered_vertex);
                fprintf(stderr, "bipartition A: ");
                for(i = 0; i < partition_A_size; i++)
                    fprintf(stderr, "%d ", partition_A[i]);
                fprintf(stderr, "\n");

                fprintf(stderr, "bipartition B: ");
                for(i = 0; i < partition_B_size; i++)
                    fprintf(stderr, "%d ", partition_B[i]);
                fprintf(stderr, "\n");
                fprintf(stderr, "\n");
            }
        }
        
        for(i = current_index; i < neighbourhood_size; i++) {
            //i.e. neighbourhood[i] is not a neighbour of a vertex which was already in a
            //If the current partition A is not valid, all super-partitions also won't be valid
            if((BIT(neighbourhood[i]) & forbidden_vertices_partition_A) == 0) {
                partition_A[partition_A_size] = neighbourhood[i];
                unsigned long long int partition_A_bitvector_new = partition_A_bitvector | BIT(partition_A[partition_A_size]);
                unsigned long long int forbidden_vertices_partition_A_new = 
                    forbidden_vertices_partition_A | neighbourhood_bitvector[partition_A[partition_A_size]];
                make_all_possible_bipartitions(partition_A, partition_A_size + 1, 
                        partition_A_bitvector_new,
                        forbidden_vertices_partition_A_new,
                        neighbourhood, neighbourhood_size, i + 1,
                        forbidden_neighbourhood_vertices);
            }
        }
    }
}

/******************************************************************************/

/**
 * Compute all bipartitions of N(vertex).
 * 
 * Note that vertex does not necessarily have a bipartite neighbourhood.
 * A vertex v has a bipartite neighbourhood if the graph induced by N(v) is bipartite.
 * 
 * If the graph induced by N(v) is not connected, it can have multiple bipartitions.
 */
static void 
compute_neighbourhood_bipartitions(unsigned char vertex) {
    //Test if graph induced by N(v) is bipartite
    //Nearly all of them are bipartite, so don't waste time testing that here (will be tested later)
    
    considered_vertex = vertex;
    unsigned long long int forbidden_vertices = neighbourhood_bitvector[vertex] | BIT(vertex);
    
    number_of_bipartitions = 0;
    number_of_reachable_bipartitions = 0;
    
    unsigned char partition_A[adj[vertex]];
    
    //Force first vertex to be in partition A to avoid duplicates
    int partition_A_size = 0;
    partition_A[partition_A_size++] = graph[vertex][0];
    unsigned long long int partition_A_bitvector = BIT(partition_A[0]);
    
    //printgraph(graph, adj, number_of_vertices);
    
    make_all_possible_bipartitions(partition_A, partition_A_size, partition_A_bitvector,
        neighbourhood_bitvector[partition_A[0]], graph[vertex], adj[vertex], 1,
            forbidden_vertices);
    
    //Max number of partitions is 2^(deg) - 1
    fprintf(stderr, "Results for vertex %d with deg %d: num of bip: %d, num of reachable bip: %d\n",
            vertex, adj[vertex], number_of_bipartitions, number_of_reachable_bipartitions);
    fprintf(stderr, "\n");
    
    vertex_found_where_property_doesnt_hold = (number_of_bipartitions > 0) 
            && (number_of_reachable_bipartitions == 0);
}

/******************************************************************************/

/**
 * Returns 1 if the inputgraph has the desired property (i.e. that Assumption 1 of Theorem 4.6 of [1] holds),
 * else returns 0.
 */
static int every_bipartite_neighbourhood_vertex_has_property() {
    //Compute bitvectors with the neighbourhood of each vertex
    int i, j;
    for(i = 0; i < number_of_vertices; i++) {
        neighbourhood_bitvector[i] = 0;
        for(j = 0; j < adj[i]; j++)
            neighbourhood_bitvector[i] |= BIT(graph[i][j]);
    }    
    
    int vertex_found_where_property_doesnt_hold_all = 0;
    
    for(i = 0; i < number_of_vertices; i++) {
        vertex_found_where_property_doesnt_hold = 0;
        
        compute_neighbourhood_bipartitions(i);
        
        if(vertex_found_where_property_doesnt_hold) {
            fprintf(stderr, "Info: Found graph where property doesn't hold!\n");
            vertex_found_where_property_doesnt_hold_all = 1;
            //exit(1);
            //return 0;
        }
    }
    
    return !vertex_found_where_property_doesnt_hold_all;
}

/******************************************************************************/

/*
 *
 */
int main(int argc, char** argv) {

    
    //For the timing
    struct tms TMS;

    int codelength = 0;
    unsigned char code[MAXN * (MAXN - 1) / 2 + MAXN];
    unsigned long long int number_of_graphs_read = 0;
    unsigned long long int number_of_graphs_written = 0;

    if(sizeof (unsigned long long int)*8 < MAXN) {
        fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
        exit(1);
    }

    int nuller;
    for(; fread(code, sizeof (unsigned char), 1, stdin);) {
        number_of_vertices = code[0];
        if(number_of_vertices > MAXN) {
            fprintf(stderr, "Number of vertices is too big (limit is %d) \n", MAXN);
            exit(0);
        }
        if(code[0] == 0)
            nuller = 1;
        else
            nuller = 0;
        codelength = 1;
        while(nuller < number_of_vertices - 1) {
            code[codelength] = getc(stdin);
            if(code[codelength] == 0) nuller++;
            codelength++;
        }

        number_of_graphs_read++;
        decode_multicode(code, codelength);
        
        fprintf(stderr, "\n------------------------------------------------\n\n");
        fprintf(stderr, "Graph %llu\n", number_of_graphs_read);
        printgraph(graph, adj, number_of_vertices);
        
        if(every_bipartite_neighbourhood_vertex_has_property()) {
            if(fwrite(code, sizeof (unsigned char), codelength, stdout) < codelength) {
                fprintf(stderr, "Error: couldn't write code!\n");
                exit(1);
            }
            number_of_graphs_written++;
        }
    }
    
    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;

    fprintf(stderr, "Read %lld graphs\n", number_of_graphs_read);
    fprintf(stderr, "Written %lld graphs where property holds\n", number_of_graphs_written);
    fprintf(stderr, "CPU time filter: %.1f seconds.\n", (double) savetime / (double) CLK_TCK);

    return (EXIT_SUCCESS);
}